import { Component } from 'react';
import { observable, observe, unobserve } from '@nx-js/observer-util';

// do not bind these, they should not be directly invoked or passed as callbacks by the user
var reactInternals = new Set(['constructor', 'render', 'componentWillMount', 'componentDidMount', 'componentWillReceiveProps', 'shouldComponentUpdate', 'componentWillUpdate', 'componentDidUpdate', 'componentWillUnmount']);

// bind the methods from proto to the passed context object and assign them to the context
function autoBind(context, proto, isReact) {
  for (var i = 0, list = Object.getOwnPropertyNames(proto); i < list.length; i += 1) {
    var key = list[i];

    var ref = Object.getOwnPropertyDescriptor(proto, key);
    var value = ref.value;

    // do not try to bind properties and getter/setters
    if (typeof value === 'function' && !(isReact && reactInternals.has(key))) {
      context[key] = value.bind(context);
    }
  }
}

function view(Comp) {
  if (typeof Comp !== 'function') {
    throw new TypeError('view() expects a component as argument.');
  }
  // wrap the component in a reactive HOC
  return toReactiveComp(Comp);
}

function toReactiveComp(Comp) {
  var isStatelessComp = !(Comp.prototype && Comp.prototype.isReactComponent);
  var BaseComp = isStatelessComp ? Component : Comp;
  // return a HOC which overwrites render, shouldComponentUpdate and componentWillUnmount
  // it decides when to run the new reactive methods and when to proxy to the original methods
  var ReactiveHOC = (function (BaseComp) {
    function ReactiveHOC(props, context) {
      var this$1 = this;

      BaseComp.call(this, props, context);

      this.state = {
        counter: 0
      };
      if (!isStatelessComp) {
        // auto bind non react specific original methods to the component instance
        autoBind(this, Comp.prototype, true);
      }

      // create a reactive render for the component
      this.render = observe(this.render, {
        scheduler: function () { return this$1.setState({ counter: this$1.state.counter + 1 }); },
        lazy: true
      });
    }

    if ( BaseComp ) ReactiveHOC.__proto__ = BaseComp;
    ReactiveHOC.prototype = Object.create( BaseComp && BaseComp.prototype );
    ReactiveHOC.prototype.constructor = ReactiveHOC;

    ReactiveHOC.prototype.render = function render () {
      return isStatelessComp ? Comp(this.props, this.context) : BaseComp.prototype.render.call(this);
    };

    // react should trigger updates on prop changes, while easyState handles store changes
    ReactiveHOC.prototype.shouldComponentUpdate = function shouldComponentUpdate (nextProps, nextState) {
      var ref = this;
      var props = ref.props;
      var state = ref.state;

      // respect the case when user prohibits updates
      if (BaseComp.prototype.shouldComponentUpdate && !BaseComp.prototype.shouldComponentUpdate.call(this, nextProps, nextState)) {
        return false;
      }

      // return true if it is a reactive render
      if (state.counter !== nextState.counter) {
        return true;
      }

      // the component should update if any of its props shallowly changed value
      var keys = Object.keys(props);
      var nextKeys = Object.keys(nextProps);
      return nextKeys.length !== keys.length || nextKeys.some(function (key) { return props[key] !== nextProps[key]; });
    };

    ReactiveHOC.prototype.componentWillUnmount = function componentWillUnmount () {
      // call user defined componentWillUnmount
      if (BaseComp.prototype.componentWillUnmount) {
        BaseComp.prototype.componentWillUnmount.call(this);
      }
      // clean up memory used by easyState
      unobserve(this.render);
    };

    return ReactiveHOC;
  }(BaseComp));
  // proxy react specific static variables to the reactive component
  copyStaticProps(Comp, ReactiveHOC);
  return ReactiveHOC;
}

// copy react specific static props between passed and HOC components
function copyStaticProps(fromComp, toComp) {
  toComp.displayName = fromComp.displayName || fromComp.name;
  toComp.contextTypes = fromComp.contextTypes;
  toComp.childContextTypes = fromComp.childContextTypes;
  toComp.propTypes = fromComp.propTypes;
  toComp.defaultProps = fromComp.defaultProps;
}

function store(obj) {
  if ( obj === void 0 ) obj = {};

  if (typeof obj !== 'object' || obj === null) {
    throw new TypeError('store() expects an object or undefined as argument');
  }
  // create an observable store from the passed object
  // and binds all of its methods to the created observable
  var observableStore = observable(obj);
  autoBind(observableStore, obj, false);
  return observableStore;
}

export { view, store };
