import { Component } from 'react';
import { observable, observe, unobserve } from '@nx-js/observer-util';

// do not bind these, they should not be directly invoked or passed as callbacks by the user
const reactInternals = new Set(['constructor', 'render', 'componentWillMount', 'componentDidMount', 'componentWillReceiveProps', 'shouldComponentUpdate', 'componentWillUpdate', 'componentDidUpdate', 'componentWillUnmount']);

// bind the methods from proto to the passed context object and assign them to the context
function autoBind(context, proto, isReact) {
  for (let key of Object.getOwnPropertyNames(proto)) {
    const { value } = Object.getOwnPropertyDescriptor(proto, key);

    // do not try to bind properties and getter/setters
    if (typeof value === 'function' && !(isReact && reactInternals.has(key))) {
      context[key] = value.bind(context);
    }
  }
}

function view(Comp) {
  if (typeof Comp !== 'function') {
    throw new TypeError('view() expects a component as argument.');
  }
  // wrap the component in a reactive HOC
  return toReactiveComp(Comp);
}

function toReactiveComp(Comp) {
  const isStatelessComp = !(Comp.prototype && Comp.prototype.isReactComponent);
  const BaseComp = isStatelessComp ? Component : Comp;
  // return a HOC which overwrites render, shouldComponentUpdate and componentWillUnmount
  // it decides when to run the new reactive methods and when to proxy to the original methods
  class ReactiveHOC extends BaseComp {

    constructor(props, context) {
      super(props, context);

      this.state = {
        counter: 0
      };
      if (!isStatelessComp) {
        // auto bind non react specific original methods to the component instance
        autoBind(this, Comp.prototype, true);
      }

      // create a reactive render for the component
      this.render = observe(this.render, {
        scheduler: () => this.setState({ counter: this.state.counter + 1 }),
        lazy: true
      });
    }

    render() {
      return isStatelessComp ? Comp(this.props, this.context) : super.render();
    }

    // react should trigger updates on prop changes, while easyState handles store changes
    shouldComponentUpdate(nextProps, nextState) {
      const { props, state } = this;

      // respect the case when user prohibits updates
      if (super.shouldComponentUpdate && !super.shouldComponentUpdate(nextProps, nextState)) {
        return false;
      }

      // return true if it is a reactive render
      if (state.counter !== nextState.counter) {
        return true;
      }

      // the component should update if any of its props shallowly changed value
      const keys = Object.keys(props);
      const nextKeys = Object.keys(nextProps);
      return nextKeys.length !== keys.length || nextKeys.some(key => props[key] !== nextProps[key]);
    }

    componentWillUnmount() {
      // call user defined componentWillUnmount
      if (super.componentWillUnmount) {
        super.componentWillUnmount();
      }
      // clean up memory used by easyState
      unobserve(this.render);
    }
  }
  // proxy react specific static variables to the reactive component
  copyStaticProps(Comp, ReactiveHOC);
  return ReactiveHOC;
}

// copy react specific static props between passed and HOC components
function copyStaticProps(fromComp, toComp) {
  toComp.displayName = fromComp.displayName || fromComp.name;
  toComp.contextTypes = fromComp.contextTypes;
  toComp.childContextTypes = fromComp.childContextTypes;
  toComp.propTypes = fromComp.propTypes;
  toComp.defaultProps = fromComp.defaultProps;
}

function store(obj = {}) {
  if (typeof obj !== 'object' || obj === null) {
    throw new TypeError('store() expects an object or undefined as argument');
  }
  // create an observable store from the passed object
  // and binds all of its methods to the created observable
  const observableStore = observable(obj);
  autoBind(observableStore, obj, false);
  return observableStore;
}

export { view, store };
